@@ -50,6 +50,18 @@
 /**                                                                         **/
 /*****************************************************************************/
 
+// Type pun to enable drawing 'long long' at random -- unif_rand() is only
+//   suitable for drawing 32-bit values, so we take the approach of drawing
+//   two such values and concatenating them. The way this works is that the
+//   LongLongRepr is exactly 2^32*high + low.
+typedef union {
+  struct {
+    unsigned int low;
+    unsigned int high;
+  } U32x2Repr;
+  long long LongLongRepr;
+} PunnedU32x2AndLongLong;
+
 /*****************************************************************************/
 /**                                                                         **/
 /**                   PROTOTYPYPES OF LOCAL FUNCTIONS                       **/
@@ -92,15 +104,15 @@ SEXP as_integer64_double(SEXP x_, SEXP ret_){
   double imax = (double) MAX_INTEGER64;
   Rboolean naflag = FALSE;
   for (i=0; i<n; i++){
-    if (ISNAN(x[i])) 
-	  ret[i] = NA_INTEGER64;
-	else{
-	  if (x[i]<imin || x[i]>imax){
-	    ret[i] = NA_INTEGER64;
-		naflag = TRUE;
-	  }else
+    if (ISNAN(x[i]))
+      ret[i] = NA_INTEGER64;
+    else{
+      if (x[i]<imin || x[i]>imax){
+        ret[i] = NA_INTEGER64;
+        naflag = TRUE;
+      }else
         ret[i] = (long long) x[i];
-	}
+    }
   }
   if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
@@ -109,11 +121,11 @@ SEXP as_integer64_double(SEXP x_, SEXP ret_){
 SEXP as_integer64_integer(SEXP x_, SEXP ret_){
   long long i, n = LENGTH(x_);
   long long * ret = (long long *) REAL(ret_);
-  int * x = INTEGER(x_); 
+  int * x = INTEGER(x_);
   for (i=0; i<n; i++){
-    if (x[i]==NA_INTEGER) 
-	  ret[i] = NA_INTEGER64;
-	else
+    if (x[i]==NA_INTEGER)
+      ret[i] = NA_INTEGER64;
+    else
       ret[i] = (long long) x[i];
   }
   return ret_;
@@ -122,7 +134,7 @@ SEXP as_integer64_integer(SEXP x_, SEXP ret_){
 
 SEXP as_double_integer64(SEXP x_, SEXP ret_){
   long long i, n = LENGTH(x_);
-  long long * x = (long long *) REAL(x_); 
+  long long * x = (long long *) REAL(x_);
   double * ret = REAL(ret_);
   double rmax = pow(FLT_RADIX, DBL_MANT_DIG) - 1;
   double rmin = -rmax;
@@ -130,11 +142,11 @@ SEXP as_double_integer64(SEXP x_, SEXP ret_){
   for (i=0; i<n; i++){
     if (x[i]==NA_INTEGER64)
       ret[i] = NA_REAL;
-	else{
-	  if (x[i]<rmin || x[i]>rmax)
-		naflag = TRUE;
-	  ret[i] = (double) x[i];
-	}
+    else{
+      if (x[i]<rmin || x[i]>rmax)
+        naflag = TRUE;
+      ret[i] = (double) x[i];
+    }
   }
   if (naflag)warning(INTEGER64_TODOUBLE_WARNING);
   return ret_;
@@ -142,19 +154,19 @@ SEXP as_double_integer64(SEXP x_, SEXP ret_){
 
 SEXP as_integer_integer64(SEXP x_, SEXP ret_){
   long long i, n = LENGTH(x_);
-  long long * x = (long long *) REAL(x_); 
+  long long * x = (long long *) REAL(x_);
   int * ret = INTEGER(ret_);
   Rboolean naflag = FALSE;
   for (i=0; i<n; i++){
     if (x[i]==NA_INTEGER64)
       ret[i] = NA_INTEGER;
-	else{
-	  if (x[i]<MIN_INTEGER32 || x[i]>MAX_INTEGER32){
-	    ret[i] = NA_INTEGER;
-		naflag = TRUE;
-	  }else
-	    ret[i] = (int) x[i];
-	}
+    else{
+      if (x[i]<MIN_INTEGER32 || x[i]>MAX_INTEGER32){
+        ret[i] = NA_INTEGER;
+        naflag = TRUE;
+      }else
+        ret[i] = (int) x[i];
+    }
   }
   if (naflag)warning(INTEGER32_OVERFLOW_WARNING);
   return ret_;
@@ -162,14 +174,14 @@ SEXP as_integer_integer64(SEXP x_, SEXP ret_){
 
 SEXP as_logical_integer64(SEXP x_, SEXP ret_){
   long long i, n = LENGTH(x_);
-  long long * x = (long long *) REAL(x_); 
+  long long * x = (long long *) REAL(x_);
   int * ret = INTEGER(ret_);
   for (i=0; i<n; i++){
     if (x[i]==NA_INTEGER64)
       ret[i] = NA_INTEGER;
-	else{
-	  ret[i] = x[i]==0 ? 0: 1;
-	}
+    else{
+      ret[i] = x[i]==0 ? 0: 1;
+    }
   }
   return ret_;
 }
@@ -181,11 +193,11 @@ SEXP as_character_integer64(SEXP x_, SEXP ret_){
   static char buff[NCHARS_DECS_INTEGER64];
   for(i=0; i<n; i++){
     if (x[i]==NA_INTEGER64){
-	  SET_STRING_ELT(ret_, i, NA_STRING);
-	}else{
-	  snprintf(buff, NCHARS_DECS_INTEGER64, COERCE_INTEGER64, x[i]); 
-	  SET_STRING_ELT(ret_, i, mkChar(buff)); 
-	}
+      SET_STRING_ELT(ret_, i, NA_STRING);
+    }else{
+      snprintf(buff, NCHARS_DECS_INTEGER64, COERCE_INTEGER64, x[i]);
+      SET_STRING_ELT(ret_, i, mkChar(buff));
+    }
   }
   return ret_;
 }
@@ -196,40 +208,73 @@ SEXP as_integer64_character(SEXP x_, SEXP ret_){
   const char * str;
   char * endpointer;
   for(i=0; i<n; i++){
-	str = CHAR(STRING_ELT(x_, i)); endpointer = (char *)str; // thanks to Murray Stokely 28.1.2012
-	ret[i] = strtoll(str, &endpointer, 10);
-	if (*endpointer)
-	  ret[i] = NA_INTEGER64;
+    str = CHAR(STRING_ELT(x_, i)); endpointer = (char *)str; // thanks to Murray Stokely 28.1.2012
+    ret[i] = strtoll(str, &endpointer, 10);
+    if (*endpointer)
+      ret[i] = NA_INTEGER64;
   }
   return ret_;
 }
 
 SEXP as_bitstring_integer64(SEXP x_, SEXP ret_){
-  long long i, n = LENGTH(ret_);
+  int i, n = LENGTH(ret_);
   long long * x = (long long *) REAL(x_);
   unsigned long long mask;
   long long v;
   static char buff[NCHARS_BITS_INTEGER64];
   char * str;
   for(i=0; i<n; i++){
-	v = x[i];
-	str = buff;
-	mask = LEFTBIT_INTEGER64;
+    v = x[i];
+    str = buff;
+    mask = LEFTBIT_INTEGER64;
     while (mask){
         if (v & mask)
               *str = '1';
-          else 
+          else
               *str = '0';
         str++;
         mask >>= 1;
     }
     *str = 0;
-    SET_STRING_ELT(ret_, i, mkChar(buff)); 
+    SET_STRING_ELT(ret_, i, mkChar(buff));
+    R_CheckUserInterrupt();
   }
   return ret_;
 }
 
-SEXP plus_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
+SEXP as_integer64_bitstring(SEXP x_, SEXP ret_){
+  Rboolean naflag = FALSE;
+  int i, k, l, n = LENGTH(x_);
+  long long * ret = (long long *) REAL(ret_);
+  unsigned long long mask;
+  long long v;
+  const char * str;
+  for(i=0; i<n; i++){
+    str = CHAR(STRING_ELT(x_, i));
+    l = strlen(str);
+    if (l>BITS_INTEGER64){
+      ret[i] = NA_INTEGER64;
+      naflag = TRUE;
+      break;
+    }
+    mask = 1;
+    v = 0;
+    for (k=l-1; k>=0; k--){
+      if (str[k] != '0' &&  str[k] != ' '){
+        v |= mask;
+      }
+      mask <<= 1;
+    }
+    ret[i] = v;
+    R_CheckUserInterrupt();
+  }
+  if (naflag)warning(BITSTRING_OVERFLOW_WARNING);
+  return ret_;
+}
+
+
+
+__attribute__((no_sanitize("signed-integer-overflow"))) SEXP plus_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long i, n = LENGTH(ret_);
   long long i1, n1 = LENGTH(e1_);
   long long i2, n2 = LENGTH(e2_);
@@ -237,14 +282,14 @@ SEXP plus_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e2 = (long long *) REAL(e2_);
   long long * ret = (long long *) REAL(ret_);
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		PLUS64(e1[i1],e2[i2],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        PLUS64(e1[i1],e2[i2],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
-SEXP minus_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
+__attribute__((no_sanitize("signed-integer-overflow"))) SEXP minus_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long i, n = LENGTH(ret_);
   long long i1, n1 = LENGTH(e1_);
   long long i2, n2 = LENGTH(e2_);
@@ -252,14 +297,14 @@ SEXP minus_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e2 = (long long *) REAL(e2_);
   long long * ret = (long long *) REAL(ret_);
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		MINUS64(e1[i1],e2[i2],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        MINUS64(e1[i1],e2[i2],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
-SEXP diff_integer64(SEXP x_, SEXP lag_, SEXP n_, SEXP ret_){
+__attribute__((no_sanitize("signed-integer-overflow"))) SEXP diff_integer64(SEXP x_, SEXP lag_, SEXP n_, SEXP ret_){
   long long i, n = *((long long *) REAL(n_));
   long long * x = (long long *) REAL(x_);
   long long * lag = (long long *) REAL(lag_);
@@ -267,11 +312,11 @@ SEXP diff_integer64(SEXP x_, SEXP lag_, SEXP n_, SEXP ret_){
   long long vlag = *lag;
   long long v;
   Rboolean naflag = FALSE;
-	for(i=0; i<n; i++) {
-	  v = x[i];
-	  MINUS64(x[i+vlag],v,ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+    for(i=0; i<n; i++) {
+      v = x[i];
+      MINUS64(x[i+vlag],v,ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
@@ -283,10 +328,10 @@ SEXP intdiv_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e2 = (long long *) REAL(e2_);
   long long * ret = (long long *) REAL(ret_);
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		INTDIV64(e1[i1],e2[i2],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_DIVISION_BY_ZERO_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        INTDIV64(e1[i1],e2[i2],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_DIVISION_BY_ZERO_WARNING);
   return ret_;
 }
 
@@ -298,15 +343,15 @@ SEXP mod_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e2 = (long long *) REAL(e2_);
   long long * ret = (long long *) REAL(ret_);
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		MOD64(e1[i1],e2[i2],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_DIVISION_BY_ZERO_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        MOD64(e1[i1],e2[i2],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_DIVISION_BY_ZERO_WARNING);
   return ret_;
 }
 
 
-SEXP times_integer64_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
+__attribute__((no_sanitize("signed-integer-overflow"))) SEXP times_integer64_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long i, n = LENGTH(ret_);
   long long i1, n1 = LENGTH(e1_);
   long long i2, n2 = LENGTH(e2_);
@@ -314,10 +359,10 @@ SEXP times_integer64_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e2 = (long long *) REAL(e2_);
   long long * ret = (long long *) REAL(ret_);
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		PROD64(e1[i1],e2[i2],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        PROD64(e1[i1],e2[i2],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
@@ -330,10 +375,10 @@ SEXP times_integer64_double(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   long double longret;
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		PROD64REAL(e1[i1],e2[i2],ret[i],naflag,longret)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        PROD64REAL(e1[i1],e2[i2],ret[i],naflag,longret)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
@@ -346,10 +391,10 @@ SEXP power_integer64_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   long double longret;
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		POW64(e1[i1],e2[i2],ret[i],naflag, longret)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        POW64(e1[i1],e2[i2],ret[i],naflag, longret)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
@@ -362,10 +407,10 @@ SEXP power_integer64_double(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   long double longret;
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		POW64REAL(e1[i1],e2[i2],ret[i],naflag,longret)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        POW64REAL(e1[i1],e2[i2],ret[i],naflag,longret)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
@@ -377,10 +422,10 @@ SEXP divide_integer64_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
    long long * e2 = (long long *) REAL(e2_);
    double * ret = REAL(ret_);
    Rboolean naflag = FALSE;
-	 mod_iterate(n1, n2, i1, i2) {
-		 DIVIDE64(e1[i1],e2[i2],ret[i],naflag)
-	 }
-	 if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+     mod_iterate(n1, n2, i1, i2) {
+         DIVIDE64(e1[i1],e2[i2],ret[i],naflag)
+     }
+     if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
    return ret_;
 }
 SEXP divide_integer64_double(SEXP e1_, SEXP e2_, SEXP ret_){
@@ -391,20 +436,37 @@ SEXP divide_integer64_double(SEXP e1_, SEXP e2_, SEXP ret_){
    double * e2 = REAL(e2_);
    double * ret = REAL(ret_);
    Rboolean naflag = FALSE;
-	 mod_iterate(n1, n2, i1, i2) {
-		 DIVIDE64REAL(e1[i1],e2[i2],ret[i],naflag)
-	 }
-	 if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+     mod_iterate(n1, n2, i1, i2) {
+         DIVIDE64REAL(e1[i1],e2[i2],ret[i],naflag)
+     }
+     if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
    return ret_;
 }
 
+
+/* Ofek Shilon */
+SEXP divide_double_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
+  long long i, n = LENGTH(ret_);
+  long long i1, n1 = LENGTH(e1_);
+  long long i2, n2 = LENGTH(e2_);
+  long long * e2 = (long long *) REAL(e2_);
+  double * e1 = REAL(e1_);
+  double * ret = REAL(ret_);
+  Rboolean naflag = FALSE;
+  mod_iterate(n1, n2, i1, i2) {
+    DIVIDEREAL64(e1[i1],e2[i2],ret[i],naflag)
+  }
+  if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+  return ret_;
+}
+
 SEXP sign_integer64(SEXP e1_, SEXP ret_){
   long long i, n = LENGTH(ret_);
   long long * e1 = (long long *) REAL(e1_);
   long long * ret = (long long *) REAL(ret_);
-	for(i=0; i<n; i++) {
-		SIGN64(e1[i],ret[i])
-	}
+    for(i=0; i<n; i++) {
+        SIGN64(e1[i],ret[i])
+    }
   return ret_;
 }
 
@@ -412,9 +474,9 @@ SEXP abs_integer64(SEXP e1_, SEXP ret_){
   long long i, n = LENGTH(ret_);
   long long * e1 = (long long *) REAL(e1_);
   long long * ret = (long long *) REAL(ret_);
-	for(i=0; i<n; i++) {
-		ABS64(e1[i],ret[i])
-	}
+    for(i=0; i<n; i++) {
+        ABS64(e1[i],ret[i])
+    }
   return ret_;
 }
 
@@ -424,7 +486,7 @@ SEXP sqrt_integer64(SEXP e1_, SEXP ret_){
   double * ret = REAL(ret_);
   Rboolean naflag = FALSE;
   for(i=0; i<n; i++) {
-	SQRT64(e1[i],ret[i],naflag)
+    SQRT64(e1[i],ret[i],naflag)
   }
   if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
   return ret_;
@@ -435,10 +497,10 @@ SEXP log_integer64(SEXP e1_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   double * ret = REAL(ret_);
   Rboolean naflag = FALSE;
-	for(i=0; i<n; i++) {
-		LOG64(e1[i],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
+    for(i=0; i<n; i++) {
+        LOG64(e1[i],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
   return ret_;
 }
 
@@ -450,10 +512,10 @@ SEXP logvect_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   double * e2 = REAL(e2_);
   double * ret = REAL(ret_);
   Rboolean naflag = FALSE;
-	mod_iterate(n1, n2, i1, i2) {
-		LOGVECT64(e1[i],e2[i],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
+    mod_iterate(n1, n2, i1, i2) {
+        LOGVECT64(e1[i],e2[i],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
   return ret_;
 }
 
@@ -463,10 +525,10 @@ SEXP logbase_integer64(SEXP e1_, SEXP base_, SEXP ret_){
   long double logbase = (long double) log(asReal(base_));
   double * ret = REAL(ret_);
   Rboolean naflag = (asReal(base_)>0) ? FALSE : TRUE;
-	for(i=0; i<n; i++) {
-		LOGBASE64(e1[i],logbase,ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
+    for(i=0; i<n; i++) {
+        LOGBASE64(e1[i],logbase,ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
   return ret_;
 }
 
@@ -476,15 +538,15 @@ SEXP log10_integer64(SEXP e1_, SEXP ret_){
   double * ret = REAL(ret_);
   Rboolean naflag = FALSE;
 #ifdef HAVE_LOG10
-	for(i=0; i<n; i++) {
-		LOG1064(e1[i],ret[i],naflag)
-	}
+    for(i=0; i<n; i++) {
+        LOG1064(e1[i],ret[i],naflag)
+    }
 #else
   long double logbase = (long double) log(10);
   for(i=0; i<n; i++) {
-	LOGBASE64(e1[i],logbase,ret[i],naflag)
+    LOGBASE64(e1[i],logbase,ret[i],naflag)
   }
-#endif	
+#endif
   if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
   return ret_;
 }
@@ -495,15 +557,15 @@ SEXP log2_integer64(SEXP e1_, SEXP ret_){
   double * ret = REAL(ret_);
   Rboolean naflag = FALSE;
 #ifdef HAVE_LOG2
-	for(i=0; i<n; i++) {
-		LOG264(e1[i],ret[i],naflag)
-	}
+    for(i=0; i<n; i++) {
+        LOG264(e1[i],ret[i],naflag)
+    }
 #else
   long double logbase = (long double) log(2);
   for(i=0; i<n; i++) {
-	LOGBASE64(e1[i],logbase,ret[i],naflag)
+    LOGBASE64(e1[i],logbase,ret[i],naflag)
   }
-#endif	
+#endif
   if (naflag)warning(INTEGER64_NAN_CREATED_WARNING);
   return ret_;
 }
@@ -513,25 +575,25 @@ SEXP any_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
   Rboolean hasna=FALSE;
-	if (asLogical(na_rm_)){
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64 && e1[i]){
-				ret[0] = TRUE;
-				return ret_;
-			}
-		}
-		ret[0] = FALSE;
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				hasna = TRUE;
-			}else if (e1[i]){
-				ret[0] = TRUE;
-				return ret_;
-			}
-		}
-		ret[0] = hasna ? NA_LOGICAL : FALSE;
-	}
+    if (asLogical(na_rm_)){
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64 && e1[i]){
+                ret[0] = TRUE;
+                return ret_;
+            }
+        }
+        ret[0] = FALSE;
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                hasna = TRUE;
+            }else if (e1[i]){
+                ret[0] = TRUE;
+                return ret_;
+            }
+        }
+        ret[0] = hasna ? NA_LOGICAL : FALSE;
+    }
   return ret_;
 }
 
@@ -540,25 +602,25 @@ SEXP all_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
   Rboolean hasna=FALSE;
-	if (asLogical(na_rm_)){
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64 && !e1[i]){
-				ret[0] = FALSE;
-				return ret_;
-			}
-		}
-		ret[0] = TRUE;
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				hasna = TRUE;
-			}else if (!e1[i]){
-				ret[0] = FALSE;
-				return ret_;
-			}
-		}
-		ret[0] = hasna ? NA_LOGICAL : TRUE;
-	}
+    if (asLogical(na_rm_)){
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64 && !e1[i]){
+                ret[0] = FALSE;
+                return ret_;
+            }
+        }
+        ret[0] = TRUE;
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                hasna = TRUE;
+            }else if (!e1[i]){
+                ret[0] = FALSE;
+                return ret_;
+            }
+        }
+        ret[0] = hasna ? NA_LOGICAL : TRUE;
+    }
   return ret_;
 }
 
@@ -569,63 +631,63 @@ SEXP sum_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   long long cumsum, tempsum;
   cumsum = 0;
-	if (asLogical(na_rm_)){
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64){
-				tempsum = cumsum + e1[i];
-				if (!GOODISUM64(cumsum, e1[i], tempsum)){
-					warning(INTEGER64_OVERFLOW_WARNING);
-					ret[0] = NA_INTEGER64;
-					return ret_;
-				}
-				cumsum = tempsum;
-			}
-		}
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				ret[0] = NA_INTEGER64;
-				return ret_;
-			}else{
-				tempsum = cumsum + e1[i];
-				if (!GOODISUM64(cumsum, e1[i], tempsum)){
-					warning(INTEGER64_OVERFLOW_WARNING);
-					ret[0] = NA_INTEGER64;
-					return ret_;
-				}
-				cumsum = tempsum;
-			}
-		}
-	}
+    if (asLogical(na_rm_)){
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64){
+                tempsum = cumsum + e1[i];
+                if (!GOODISUM64(cumsum, e1[i], tempsum)){
+                    warning(INTEGER64_OVERFLOW_WARNING);
+                    ret[0] = NA_INTEGER64;
+                    return ret_;
+                }
+                cumsum = tempsum;
+            }
+        }
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                ret[0] = NA_INTEGER64;
+                return ret_;
+            }else{
+                tempsum = cumsum + e1[i];
+                if (!GOODISUM64(cumsum, e1[i], tempsum)){
+                    warning(INTEGER64_OVERFLOW_WARNING);
+                    ret[0] = NA_INTEGER64;
+                    return ret_;
+                }
+                cumsum = tempsum;
+            }
+        }
+    }
   ret[0] = cumsum;
   return ret_;
 }
 
 SEXP mean_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
-	long long i, n = LENGTH(e1_);
-	long long * e1 = (long long *) REAL(e1_);
-	long long * ret = (long long *) REAL(ret_);
-	long double longret = 0;
-	if (asLogical(na_rm_)){
-		long long nvalid = 0;
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64){
-				longret += e1[i];
-				nvalid++;
-			}
-		}
-		ret[0] = longret / nvalid;
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				ret[0] = NA_INTEGER64;
-				return ret_;
-			}else{
-				longret += e1[i];			
-			}
-		}
-		ret[0] = longret / n;
-	}
+    long long i, n = LENGTH(e1_);
+    long long * e1 = (long long *) REAL(e1_);
+    long long * ret = (long long *) REAL(ret_);
+    long double longret = 0;
+    if (asLogical(na_rm_)){
+        long long nvalid = 0;
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64){
+                longret += e1[i];
+                nvalid++;
+            }
+        }
+        ret[0] = longret / nvalid;
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                ret[0] = NA_INTEGER64;
+                return ret_;
+            }else{
+                longret += e1[i];
+            }
+        }
+        ret[0] = longret / n;
+    }
   return ret_;
 }
 
@@ -635,34 +697,34 @@ SEXP prod_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   long long cumprod, tempprod;
   cumprod = 1;
-	if (asLogical(na_rm_)){
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64){
-				tempprod = cumprod * e1[i];
-				if (!GOODIPROD64(cumprod, e1[i], tempprod)){
-					warning(INTEGER64_OVERFLOW_WARNING);
-					ret[0] = NA_INTEGER64;
-					return ret_;
-				}
-				cumprod = tempprod;
-			}
-		}
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				ret[0] = NA_INTEGER64;
-				return ret_;
-			}else{
-				tempprod = cumprod * e1[i];
-				if (!GOODIPROD64(cumprod, e1[i], tempprod)){
-					warning(INTEGER64_OVERFLOW_WARNING);
-					ret[0] = NA_INTEGER64;
-					return ret_;
-				}
-				cumprod = tempprod;
-			}
-		}
-	}
+    if (asLogical(na_rm_)){
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64){
+                tempprod = cumprod * e1[i];
+                if (!GOODIPROD64(cumprod, e1[i], tempprod)){
+                    warning(INTEGER64_OVERFLOW_WARNING);
+                    ret[0] = NA_INTEGER64;
+                    return ret_;
+                }
+                cumprod = tempprod;
+            }
+        }
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                ret[0] = NA_INTEGER64;
+                return ret_;
+            }else{
+                tempprod = cumprod * e1[i];
+                if (!GOODIPROD64(cumprod, e1[i], tempprod)){
+                    warning(INTEGER64_OVERFLOW_WARNING);
+                    ret[0] = NA_INTEGER64;
+                    return ret_;
+                }
+                cumprod = tempprod;
+            }
+        }
+    }
   ret[0] = cumprod;
   return ret_;
 }
@@ -673,23 +735,23 @@ SEXP min_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * ret = (long long *) REAL(ret_);
   ret[0] = MAX_INTEGER64;
-	if (asLogical(na_rm_)){
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64 && e1[i]<ret[0]){
-				ret[0] = e1[i];
-			}
-		}
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				ret[0] = NA_INTEGER64;
-				return ret_;
-			}else{
-				if (e1[i]<ret[0])
-					ret[0] = e1[i];
-			}
-		}
-	}
+    if (asLogical(na_rm_)){
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64 && e1[i]<ret[0]){
+                ret[0] = e1[i];
+            }
+        }
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                ret[0] = NA_INTEGER64;
+                return ret_;
+            }else{
+                if (e1[i]<ret[0])
+                    ret[0] = e1[i];
+            }
+        }
+    }
   return ret_;
 }
 
@@ -698,23 +760,23 @@ SEXP max_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * ret = (long long *) REAL(ret_);
   ret[0] = MIN_INTEGER64;
-	if (asLogical(na_rm_)){
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64 && e1[i]>ret[0]){
-				ret[0] = e1[i];
-			}
-		}
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				ret[0] = NA_INTEGER64;
-				return ret_;
-			}else{
-				if (e1[i]>ret[0])
-					ret[0] = e1[i];
-			}
-		}
-	}
+    if (asLogical(na_rm_)){
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64 && e1[i]>ret[0]){
+                ret[0] = e1[i];
+            }
+        }
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                ret[0] = NA_INTEGER64;
+                return ret_;
+            }else{
+                if (e1[i]>ret[0])
+                    ret[0] = e1[i];
+            }
+        }
+    }
   return ret_;
 }
 
@@ -724,28 +786,28 @@ SEXP range_integer64(SEXP e1_, SEXP na_rm_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   ret[0] = MAX_INTEGER64;
   ret[1] = MIN_INTEGER64;
-	if (asLogical(na_rm_)){
-		for(i=0; i<n; i++){
-			if (e1[i]!=NA_INTEGER64){
-				if (e1[i]<ret[0])
-					ret[0] = e1[i];
-				if (e1[i]>ret[1])
-					ret[1] = e1[i];
-			}
-		}
-	}else{
-		for(i=0; i<n; i++){
-			if (e1[i]==NA_INTEGER64){
-				ret[0] = ret[1] = NA_INTEGER64;
-				return ret_;
-			}else{
-				if (e1[i]<ret[0])
-					ret[0] = e1[i];
-				if (e1[i]>ret[1])
-					ret[1] = e1[i];
-			}
-		}
-	}
+    if (asLogical(na_rm_)){
+        for(i=0; i<n; i++){
+            if (e1[i]!=NA_INTEGER64){
+                if (e1[i]<ret[0])
+                    ret[0] = e1[i];
+                if (e1[i]>ret[1])
+                    ret[1] = e1[i];
+            }
+        }
+    }else{
+        for(i=0; i<n; i++){
+            if (e1[i]==NA_INTEGER64){
+                ret[0] = ret[1] = NA_INTEGER64;
+                return ret_;
+            }else{
+                if (e1[i]<ret[0])
+                    ret[0] = e1[i];
+                if (e1[i]>ret[1])
+                    ret[1] = e1[i];
+            }
+        }
+    }
   return ret_;
 }
 
@@ -762,20 +824,20 @@ SEXP cummin_integer64(SEXP e1_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * ret = (long long *) REAL(ret_);
   if (n>0){
-	i=0;
-	ret[i] = e1[i];
-	if(e1[i]!=NA_INTEGER64)
-	for(i=1; i<n; i++){
-		if(e1[i]==NA_INTEGER64){
-			ret[i] = e1[i];
-			break;
-		}else{
-			ret[i] = e1[i]<ret[i-1] ? e1[i] : ret[i-1];		
-		}
-	}
-	for(i++; i<n; i++){
-		ret[i] = NA_INTEGER64;
-	}
+    i=0;
+    ret[i] = e1[i];
+    if(e1[i]!=NA_INTEGER64)
+    for(i=1; i<n; i++){
+        if(e1[i]==NA_INTEGER64){
+            ret[i] = e1[i];
+            break;
+        }else{
+            ret[i] = e1[i]<ret[i-1] ? e1[i] : ret[i-1];
+        }
+    }
+    for(i++; i<n; i++){
+        ret[i] = NA_INTEGER64;
+    }
   }
   return ret_;
 }
@@ -785,20 +847,20 @@ SEXP cummax_integer64(SEXP e1_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * ret = (long long *) REAL(ret_);
   if (n>0){
-	i=0;
-	ret[i] = e1[i];
-	if(e1[i]!=NA_INTEGER64)
-	for(i=1; i<n; i++){
-		if(e1[i]==NA_INTEGER64){
-			ret[i] = e1[i];
-			break;
-		}else{
-			ret[i] = e1[i]>ret[i-1] ? e1[i] : ret[i-1];		
-		}
-	}
-	for(i++; i<n; i++){
-		ret[i] = NA_INTEGER64;
-	}
+    i=0;
+    ret[i] = e1[i];
+    if(e1[i]!=NA_INTEGER64)
+    for(i=1; i<n; i++){
+        if(e1[i]==NA_INTEGER64){
+            ret[i] = e1[i];
+            break;
+        }else{
+            ret[i] = e1[i]>ret[i-1] ? e1[i] : ret[i-1];
+        }
+    }
+    for(i++; i<n; i++){
+        ret[i] = NA_INTEGER64;
+    }
   }
   return ret_;
 }
@@ -809,11 +871,11 @@ SEXP cumsum_integer64(SEXP e1_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   Rboolean naflag = FALSE;
     if (n>0)
-	  ret[0] = e1[0];
-	for(i=1; i<n; i++) {
-		PLUS64(e1[i],ret[i-1],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+      ret[0] = e1[0];
+    for(i=1; i<n; i++) {
+        PLUS64(e1[i],ret[i-1],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
@@ -823,11 +885,11 @@ SEXP cumprod_integer64(SEXP e1_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   Rboolean naflag = FALSE;
     if (n>0)
-	  ret[0] = e1[0];
-	for(i=1; i<n; i++) {
-		PROD64(e1[i],ret[i-1],ret[i],naflag)
-	}
-	if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
+      ret[0] = e1[0];
+    for(i=1; i<n; i++) {
+        PROD64(e1[i],ret[i-1],ret[i],naflag)
+    }
+    if (naflag)warning(INTEGER64_OVERFLOW_WARNING);
   return ret_;
 }
 
@@ -839,20 +901,20 @@ SEXP seq_integer64(SEXP from_, SEXP by_, SEXP ret_){
   long long * ret = (long long *) REAL(ret_);
   if (n>0){
     ret[0] = from[0];
-	for(i=1; i<n; i++){
-		ret[i] = ret[i-1] + by;
-	}
+    for(i=1; i<n; i++){
+        ret[i] = ret[i-1] + by;
+    }
   }
   return ret_;
 }
 
 SEXP isna_integer64(SEXP e1_, SEXP ret_){
-  long long i, n = LENGTH(ret_);
+  long long i, n = XLENGTH(ret_);
   long long * e1 = (long long *) REAL(e1_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
-	for(i=0; i<n; i++) {
-		ret[i] = (e1[i]==NA_INTEGER64) ? TRUE : FALSE;
-	}
+    for(i=0; i<n; i++) {
+        ret[i] = (e1[i]==NA_INTEGER64) ? TRUE : FALSE;
+    }
   return ret_;
 }
 
@@ -865,9 +927,9 @@ SEXP EQ_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * e2 = (long long *) REAL(e2_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
-	mod_iterate(n1, n2, i1, i2) {
-		EQ64(e1[i1],e2[i2],ret[i])
-	}
+    mod_iterate(n1, n2, i1, i2) {
+        EQ64(e1[i1],e2[i2],ret[i])
+    }
   return ret_;
 }
 
@@ -878,9 +940,9 @@ SEXP NE_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * e2 = (long long *) REAL(e2_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
-	mod_iterate(n1, n2, i1, i2) {
-		NE64(e1[i1],e2[i2],ret[i])
-	}
+    mod_iterate(n1, n2, i1, i2) {
+        NE64(e1[i1],e2[i2],ret[i])
+    }
   return ret_;
 }
 
@@ -891,9 +953,9 @@ SEXP LT_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * e2 = (long long *) REAL(e2_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
-	mod_iterate(n1, n2, i1, i2) {
-		LT64(e1[i1],e2[i2],ret[i])
-	}
+    mod_iterate(n1, n2, i1, i2) {
+        LT64(e1[i1],e2[i2],ret[i])
+    }
   return ret_;
 }
 
@@ -904,9 +966,9 @@ SEXP LE_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * e2 = (long long *) REAL(e2_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
-	mod_iterate(n1, n2, i1, i2) {
-		LE64(e1[i1],e2[i2],ret[i])
-	}
+    mod_iterate(n1, n2, i1, i2) {
+        LE64(e1[i1],e2[i2],ret[i])
+    }
   return ret_;
 }
 
@@ -917,9 +979,9 @@ SEXP GT_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * e2 = (long long *) REAL(e2_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
-	mod_iterate(n1, n2, i1, i2) {
-		GT64(e1[i1],e2[i2],ret[i])
-	}
+    mod_iterate(n1, n2, i1, i2) {
+        GT64(e1[i1],e2[i2],ret[i])
+    }
   return ret_;
 }
 
@@ -930,12 +992,86 @@ SEXP GE_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
   long long * e1 = (long long *) REAL(e1_);
   long long * e2 = (long long *) REAL(e2_);
   Rboolean * ret = (Rboolean *) LOGICAL(ret_);
-	mod_iterate(n1, n2, i1, i2) {
-		GE64(e1[i1],e2[i2],ret[i])
-	}
+    mod_iterate(n1, n2, i1, i2) {
+        GE64(e1[i1],e2[i2],ret[i])
+    }
   return ret_;
 }
 
+void draw_u32_twice(PunnedU32x2AndLongLong *x) {
+  x->U32x2Repr.low = (unsigned int) floor(unif_rand()*4294967296 /* =2^32 */);
+  x->U32x2Repr.high = (unsigned int) floor(unif_rand()*4294967296);
+}
+
+SEXP runif_integer64(SEXP n_, SEXP min_, SEXP max_){
+  int i, n=asInteger(n_);
+  long long min = *((long long * ) REAL(min_));
+  long long max = *((long long * ) REAL(max_));
+  unsigned long long d;
+  // max - min can overflow
+  if (min < 0 && max > 0){
+     d = ((unsigned long long)(-min)) + ((unsigned long long)max) + 1;
+  }else{
+    d = (max - min) + 1;
+  }
+  SEXP ret_;
+  PROTECT(ret_ = allocVector(REALSXP, n));
+  long long * ret = (long long *) REAL(ret_);
+  PunnedU32x2AndLongLong rand_draw;
+  GetRNGstate();
+  for (i=0; i<n; i++){
+    draw_u32_twice(&rand_draw);
+    // Requires exceedingly rare (2^(-64) probability) occurrence.
+    //   In principle coverage could be done by finding with the 'perfect' random seed,
+    //   but it's not worth burning compute to find out it is.
+    // On a "normal" machine, drawing NA_INTEGER64 corresponds to low=0, high=2^31,
+    //   so in principal we could just try and steer draw_u32_twice away from this.
+    //   But the transformation becomes non-trivial, the statistics messy, and it's by no means
+    //   guaranteed to work across all platforms -- simpler to just leave this simple retry approach.
+    while(rand_draw.LongLongRepr == NA_INTEGER64)
+      draw_u32_twice(&rand_draw); // # nocov
+    ret[i] = min + (rand_draw.LongLongRepr % d);
+  }
+  PutRNGstate();
+  UNPROTECT(1);
+  return ret_;
+}
+
+
+SEXP as_list_integer64(SEXP x_){
+  long long i, n = LENGTH(x_);
+  if (n){
+    SEXP class;
+    for (i=0; i<n; i++){
+      PROTECT(class = allocVector(STRSXP, 1));
+      SET_STRING_ELT(class, 0, mkChar("integer64"));
+      classgets(VECTOR_ELT(x_, i), class);
+    }
+    UNPROTECT(n);
+  }
+  return x_;
+}
+
+
+
+
+
+
+
+
+/*
+require(bit64)
+require(microbenchmark)
+microbenchmark(runif64(1e6))
+
+sort(runif64(1e2))
+
+
+Unit: milliseconds
+           expr      min       lq     mean   median       uq      max neval
+ runif64(1e+06) 24.62306 25.60286 25.61903 25.61369 25.62032 26.40202   100
+*/
+
 
 /*****************************************************************************/
 /**                                                                         **/
@@ -954,5 +1090,5 @@ SEXP GE_integer64(SEXP e1_, SEXP e2_, SEXP ret_){
 
 
 
-	
+
 
